import logging
import re
from subprocess import PIPE
from typing import List, Tuple

import click
from delfino.decorators import pass_args
from delfino.execution import OnError, run

from delfino_core.utils import ask, assert_executable_installed

_LOGGER = logging.getLogger(__name__)
_INVALID_BRANCH_NAME_CHARS = re.compile(r"[^a-zA-Z0-9_-]+")


def _consume_args_until_next_option(passed_args: List[str]) -> Tuple[str, List[str]]:
    for index, arg in enumerate(passed_args):
        if arg.startswith("-"):
            return " ".join(passed_args[:index]), passed_args[index:]
    return " ".join(passed_args), []


def _sanitize_branch_name(branch_name: str) -> str:
    return _INVALID_BRANCH_NAME_CHARS.sub("_", branch_name.lower()).rstrip("_")


def _get_sanitized_used_name() -> str:
    # Get the current username from git config
    username = run(["git", "config", "user.email"], stdout=PIPE, on_error=OnError.ABORT).stdout.decode().strip()

    if not username:  # If not available, use the system username
        _LOGGER.warning("No git user.email found, using system username.")
        assert_executable_installed("whoami", required_by="gh")
        username = run(["whoami"], stdout=PIPE, on_error=OnError.ABORT).stdout.decode().strip()
    elif "@" in username:  # Strip the domain from the email address
        username = username.split("@")[0]

    # Sanitize the username
    return _sanitize_branch_name(username)


def _get_new_branch_name_or_switch_to_branch(title: str) -> Tuple[str, bool]:
    branch_prefix = f"{_get_sanitized_used_name()}/"
    branch_name = branch_prefix + _sanitize_branch_name(title)

    # Check if not already on the branch
    branch_exists = (
        run(["git", "branch", "--list", branch_name], stdout=PIPE, on_error=OnError.PASS).stdout.decode().strip()
    )

    if branch_exists:
        if branch_exists.startswith("* "):
            _LOGGER.warning(f"Already on branch '{branch_name}'.")
        else:
            _LOGGER.warning(f"Branch '{branch_name}' already exists, switching to it.")
            run(["git", "checkout", branch_name], on_error=OnError.ABORT)
    else:
        # Ask user if they want to use the branch name or let them choose a different one
        while not ask(f"Use branch name '{branch_name}'?"):
            new_branch_name = input(f"Enter a branch name: {branch_prefix}")
            sanitized_new_branch_name = _sanitize_branch_name(new_branch_name)
            if sanitized_new_branch_name == new_branch_name:  # not sanitized, use it
                branch_name = branch_prefix + new_branch_name
                break

    return branch_name, not branch_exists


@click.group("gh")
def run_gh():
    """Extends `gh` or passes through.

    If a `gh` sub-command is not extended, it will be passed through to the original `gh` command.

    See the help of the sub-commands for more details. See also https://cli.github.com/manual/ or `gh --help`.
    """
    assert_executable_installed("git", required_by="gh")
    assert_executable_installed("gh", required_by="gh")


@run_gh.group("pr")
def run_gh_pr():
    """Extends `gh pr` with additional functions and defaults.

    See the help of the sub-commands for more details.
    """


@run_gh_pr.command("create")
@pass_args
def run_gh_pr_create(passed_args: Tuple[str, ...]):
    """Extends `gh pr create` with additional functions and defaults.

    Defaults to `--assignee @me --base main --draft` and adds the title if provided.
    Title doesn't need to be quoted. Any word after the command that doesn't start with
    `-` is assumed to be the title.
    """
    title = None
    args = list(passed_args)
    if args and not args[0].startswith("-"):
        # First arg is not an option, assume it's the PR title
        title, args = _consume_args_until_next_option(args)

    run(
        [
            "gh",
            "pr",
            "create",
            "--assignee",
            "@me",
            "--base",
            "main",
            "--draft",
            "--body",
            "",
            *(["--title", title] if title else []),
            *args,
        ],
        on_error=OnError.ABORT,
    )


@run_gh_pr.command("start")
@pass_args
@click.pass_context
def run_gh_pr_start(click_context: click.Context, passed_args: Tuple[str, ...]):
    """Like `gh pr create`, but with additional functionality.

    - Asks for a title if not provided.
    - Stashes any uncommitted changes.
    - Switches to `main` first and pulls the latest changes.
    - Creates a new branch using git/system username and the title.
    - Creates an empty commit to be able to create a PR.
    - Push the branch and the commit to remote.
    - Pops the stash.
    """
    args = list(passed_args)
    title, args = _consume_args_until_next_option(args)
    while not title:
        title = input("Enter a title for the PR: ")

    branch_name, create_branch = _get_new_branch_name_or_switch_to_branch(title)

    if create_branch:
        run(["git", "stash"], on_error=OnError.ABORT)
        run(["git", "checkout", "main"], on_error=OnError.ABORT)
        run(["git", "pull"], on_error=OnError.ABORT)
        run(["git", "checkout", "-b", branch_name], on_error=OnError.ABORT)

        # Check if last commit messages is not title
        last_commit_message = run(
            ["git", "log", "-1", "--pretty=%B"], stdout=PIPE, on_error=OnError.ABORT
        ).stdout.decode()
        if not last_commit_message.startswith(title):
            # Create and empty commit to be able to create a PR
            run(["git", "commit", "--allow-empty", "-m", title], on_error=OnError.ABORT)

        run(["git", "push", "--set-upstream", "origin", branch_name], on_error=OnError.ABORT)
        run(["git", "stash", "pop"], on_error=OnError.ABORT)

    # Create the PR using gh
    click_context.forward(run_gh_pr_create, passed_args=["--title", title, *args])


@run_gh_pr.command("view")
@pass_args
def run_gh_pr_view(passed_args: Tuple[str, ...]):
    """Extends `gh pr view` with additional functions and defaults.

    Defaults to `--web`.
    """
    run(["gh", "pr", "view", "--web", *passed_args], on_error=OnError.ABORT)



from __future__ import annotations

import time
from dataclasses import dataclass
from itertools import cycle
from subprocess import CompletedProcess

import click


def _spinner(chars: str = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"):
    """A spinner for showing progress."""
    frames = cycle(chars)
    while True:
        yield next(frames)


def _timer():
    """A timer which shows total time elapsed on each iteration."""
    start = time.monotonic()
    while True:
        yield f"{time.monotonic() - start:.1f}s"


@dataclass(frozen=True)
class Style:
    BOLD: str = "\033[1m"
    RESET: str = "\033[0m"


@dataclass(frozen=True)
class Color:
    RED: str = "\033[31m"
    GREEN: str = "\033[32m"
    YELLOW: str = "\033[33m"
    BLUE: str = "\033[34m"
    MAGENTA: str = "\033[35m"
    CYAN: str = "\033[36m"
    WHITE: str = "\033[37m"
    BLACK: str = "\033[30m"


@dataclass(frozen=True)
class BoldColor:
    RED: str = "\033[1;31m"
    GREEN: str = "\033[1;32m"
    YELLOW: str = "\033[1;33m"
    BLUE: str = "\033[1;34m"
    MAGENTA: str = "\033[1;35m"
    CYAN: str = "\033[1;36m"
    WHITE: str = "\033[1;37m"
    BLACK: str = "\033[1;30m"


class Spinner:
    def __init__(self, name: str, description: str | None = None, fps: float = 0.05):
        self.name = name
        self.description = f": {description}" if description else ""
        self.fps = fps
        self._timer = _timer()
        self._spinner = _spinner()
        self._last_timer_msg = ""

    def __call__(self):
        self._last_timer_msg = (
            f"\r {BoldColor.BLUE}{next(self._spinner)} {self.name}{Style.RESET}{Color.BLUE} ({next(self._timer)})"
            f"{self.description}{Style.RESET}"
        )
        click.echo(self._last_timer_msg, nl=False)
        time.sleep(self.fps)

    def _result_msg(self, color: str) -> str:
        return (
            f" {Style.BOLD}{self.name}{Style.RESET}{color} ({next(self._timer)})" f"{self.description}{Style.RESET}"
        ).ljust(len(self._last_timer_msg), " ")

    def _print_failure(self, result: CompletedProcess):
        output = "\n".join(stream.decode() for stream in [result.stdout, result.stderr] if stream)
        msg = f"{self._result_msg(Color.RED)}\n\n{output}" if output else self._result_msg
        click.secho(f"\r {Color.RED}✘{msg}")

    def _print_success(self):
        click.secho(f"\r {Color.GREEN}✔{self._result_msg(Color.GREEN)}")

    def print_results(self, result: CompletedProcess, error_msg_match: str = "", error_cls: type = click.Abort):
        """Prints the result of a command run.

        The command must be run with `stdout=PIPE` and `stderr=PIPE` to capture the output
        and show it after the command has finished.

        Args:
            result: The result of the command.
            error_msg_match: A string to match against the stdout and stderr. If any of the
                streams contains this string, the command is considered to have failed.
            error_cls: The error to raise if the command failed.
        """
        if result.returncode > 0 or (
            error_msg_match and (error_msg_match in result.stdout.decode() or error_msg_match in result.stderr.decode())
        ):
            self._print_failure(result)
            raise error_cls()

        self._print_success()



from __future__ import annotations

import os
import re
import webbrowser
from datetime import datetime, timedelta
from subprocess import PIPE, CompletedProcess

import click
from click import secho
from delfino.constants import ENTRY_POINT, PackageManager
from delfino.decorators import pass_app_context
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.validation import assert_package_manager_is_known, assert_pip_package_installed

from delfino_core.commands.verify import run_group_verify
from delfino_core.config import CorePluginConfig
from delfino_core.spinner import Spinner
from delfino_core.utils import ask

try:
    from git import Repo
except ImportError:
    pass


def _run(args: str, spinner: Spinner | None = None) -> CompletedProcess:
    """Print the command before execution."""
    if spinner is None:
        return run(args, on_error=OnError.EXIT, stdout=PIPE, stderr=PIPE)

    result = run(args, on_error=OnError.PASS, running_hook=spinner, stdout=PIPE, stderr=PIPE)
    spinner.print_results(result, error_cls=click.exceptions.Exit)
    return result


class Updater:
    _FILENAME: str = ""

    @staticmethod
    def _git_root():
        return (
            run(["git", "rev-parse", "--show-toplevel"], stdout=PIPE, stderr=PIPE, on_error=OnError.EXIT)
            .stdout.decode()
            .strip()
        )

    def __init__(self, stash: bool):
        assert_pip_package_installed("gitpython")

        self._repo = Repo(self._git_root())
        self._stash = stash
        now = datetime.utcnow()
        self._start_of_week = now - timedelta(now.isoweekday() - 1)

    def commit_and_push(self):
        commit_message = f"Dependencies rollup: {self._start_of_week.strftime('%Y-%m-%d')}"
        can_update = self._repo.commit("HEAD").message.strip() == commit_message

        if do_commit := ask("Do you want to commit changes now?"):
            if can_update:  # update existing commit
                if self._repo.is_dirty():
                    _run("git pull")
                _run("git commit -a --amend -C HEAD")
            else:
                _run("git add .")
                _run(f"git commit -a -m '{commit_message}'")

        if do_commit and ask("Do you want to push changes now?"):
            if can_update:
                _run("git push --force-with-lease")
            else:
                _run("git push -u origin HEAD")

            url = self._link_to_open_a_pull_request()
            if url:
                if ask("Do you want to open a new pull request now in a web browser?"):
                    webbrowser.open(url)
                else:
                    secho(f"\nOpen a new pull request by visiting:\n\n\t{url}\n", fg="green")

    def _link_to_open_a_pull_request(self) -> str | None:
        url = self._repo.remote().url

        if not (match := re.match("git@github.com:(.*)\\.git", url)):
            match = re.match("https://github.com/(.*)\\.git", url)

        return f"https://github.com/{match.group(1)}/pull/new/{self._repo.active_branch}" if match else None

    def _read_dependency_file(self) -> str:
        with open(self._FILENAME, encoding="utf-8") as file:
            return file.read()

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        raise NotImplementedError

    def get_branch_name(self) -> str:
        email = self._repo.config_reader().get_value("user", "email")
        assert (
            email
        ), "Git config has not a commit email set. Please set it with:\n\tgit config --global user.email 'YOUR_EMAIL'"
        assert isinstance(email, str)

        user_name = re.sub("@.*", "", email)
        assert user_name, f"No user name could be parsed from git commit email '{email}' after removing '@.*'."
        assert isinstance(user_name, str)
        user_name = re.sub("[^a-zA-Z0-9._-]", "_", user_name)

        return f"{user_name}/dependencies_rollup_{self._start_of_week.strftime('%Y_%m_%d')}"

    @classmethod
    def _lock_and_sync(cls):
        pass

    def _show_edit_prompt_and_wait(self, *, available_updates: str):
        print("\n" + available_updates)
        input(
            f"\033[1;33mEdit {self._FILENAME} file to update any of the dependencies above ^^^\n\n"
            f"Then continue by pressing ENTER ...\033[0m"
        )

    def checkout_branch(self, branch: str):
        if str(self._repo.active_branch) == branch:
            secho(f"Branch '{branch}' already exists and active.", fg="green")
        elif branch in self._repo.branches:  # type: ignore[operator]
            secho(f"Branch '{branch}' already exists.", fg="yellow")
            _run(f"git checkout {branch}")
        else:
            if self._stash:
                secho("Stashing existing changes.", fg="yellow")
                _run("git stash")

            if self._repo.active_branch != "main":
                secho("Checking out 'main'.", fg="yellow")
                _run("git checkout main")

            secho("Pulling latest changes.", fg="yellow")
            _run("git pull")

            secho(f"Creating a new branch '{branch}'.", fg="yellow")
            _run(f"git checkout -b {branch}")
            self._lock_and_sync()

    def update(self, retry: bool):
        branch_name = self.get_branch_name()
        self.checkout_branch(branch_name)

        if not retry:
            while True:
                if not self.print_outdated_packages_and_lock_if_changed():
                    break

            secho("Running all tests to check the updated dependencies ... ", fg="yellow")


class PipenvUpdater(Updater):
    _FILENAME = "Pipfile"

    _SKIP_PATTERN = (
        "Skipped Update of Package (?P<package>[^:]+): (?P<installed>[^ ]+) "
        "installed.* (?P<available>[^ ]+) available.*"
    )
    _OUTDATED_PATTEN = (
        "Package '(?P<package>[^']+)' out-of-date: '(?P<installed>[^']+)' "
        "installed.* '(?P<available>[^']+)' available.*"
    )
    _VERSION_CONSTRAINT_CHARS = "=~<>"

    @classmethod
    def _lock_and_sync(cls):
        _run("pipenv lock")
        _run("pipenv sync -d")

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        spinner = Spinner("pipenv", "updating packages based on version pinning")
        _run("pipenv update -d", spinner)

        spinner = Spinner("pipenv", "checking outdated packages")
        result = _run("pipenv update --outdated", spinner)

        pipfile = self._read_dependency_file()

        available_updates = []

        for line in result.stdout.decode().split(os.linesep) + result.stderr.decode().split(os.linesep):
            if not (match := re.match(self._SKIP_PATTERN, line) or re.match(self._OUTDATED_PATTEN, line)):
                continue

            package = match.group("package")
            installed = match.group("installed").lstrip(self._VERSION_CONSTRAINT_CHARS)
            available = match.group("available").lstrip(self._VERSION_CONSTRAINT_CHARS)

            # Keep only packages defined in Pipfile with a different version available
            if installed != available and re.search(f'\n"?{package}"? ', pipfile):
                available_updates.append(f"{package}: {installed} -> {available}")

        if not available_updates:
            return False

        self._show_edit_prompt_and_wait(available_updates="\n".join(sorted(available_updates)) + "\n")

        return self._read_dependency_file() != pipfile


class PoetryUpdater(Updater):
    _FILENAME = "pyproject.toml"

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        spinner = Spinner("poetry", "updating packages based on version pinning")
        _run("poetry update", spinner)

        spinner = Spinner("poetry", "checking outdated packages")
        if not (result := _run("poetry show --outdated --why --ansi", spinner)).stdout:
            return False

        pyproject_toml = self._read_dependency_file()

        self._show_edit_prompt_and_wait(available_updates=result.stdout.decode())

        return self._read_dependency_file() != pyproject_toml


@click.command("dependencies-update")
@click.option("--retry", default=False, show_default=True, is_flag=True, help="Retry an update after failed tests.")
@click.option(
    "--stash/--no-stash",
    default=True,
    show_default=True,
    help=(
        "Don't stash existing changes. Useful when there are some existing manual changes "
        "that should be also part of the upgrade."
    ),
)
@pass_app_context(CorePluginConfig)
@click.pass_context
def run_dependencies_update(click_context: click.Context, app_context: AppContext, retry: bool, stash: bool):
    """Manages the process of updating dependencies."""
    print_header("Updating dependencies", icon="🔄")

    assert_package_manager_is_known(app_context.package_manager)
    updater: Updater

    if app_context.package_manager == PackageManager.PIPENV:
        updater = PipenvUpdater(stash)
    elif app_context.package_manager == PackageManager.POETRY:
        updater = PoetryUpdater(stash)
    else:
        raise AssertionError(
            f"The '{app_context.package_manager.value}' package manager is not supported by this command."
        )

    updater.update(retry)

    try:
        click_context.invoke(run_group_verify)
    except Exception:
        secho(
            f"\nOne or more checks have failed. Fix any issues above and then run:\n\n\t"
            f"{ENTRY_POINT} {click_context.info_name} --retry\n",
            fg="red",
        )
        raise

    updater.commit_and_push()



import click
from delfino.models import AppContext

from delfino_core.commands.format import run_group_format
from delfino_core.commands.lint import run_group_lint
from delfino_core.commands.test import run_group_test
from delfino_core.commands.typecheck import run_mypy
from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import commands_group_help, execute_commands_group

_COMMANDS = [run_group_format, run_group_lint, run_mypy, run_group_test]


@click.command("verify", help=commands_group_help("verify"))
@pass_plugin_app_context
@click.pass_context
def run_group_verify(click_context: click.Context, app_context: AppContext[CorePluginConfig]):
    execute_commands_group(click_context, app_context.plugin_config)



from pathlib import Path
from subprocess import PIPE, CompletedProcess
from typing import List, Optional, Tuple

import click
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import run_command_example
from delfino.validation import assert_pip_package_installed
from packaging.specifiers import SpecifierSet

from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.spinner import Spinner
from delfino_core.utils import commands_group_help, execute_commands_group


def _major_minor_only(version_number: str) -> str:
    return ".".join(version_number.split(".")[:2])


def _find_min_minor_version(version_spec: str, major: int, minor_min: int, minor_max: int) -> Optional[str]:
    version_spec = version_spec.replace("^", ">=")

    # Drop patch and lower version number parts because they are not relevant.
    # For example, "3.8" does not fit "^3.8.1" and "3.9" would be mistakenly used.
    version_spec = ",".join(_major_minor_only(spec_part) for spec_part in version_spec.split(","))

    for minor in range(minor_min, minor_max + 1):
        version = f"{major}.{minor}"
        if SpecifierSet(version_spec).contains(version):
            return str(minor)

    return ""


def _all_python_files(files_folders: Tuple[Path, ...]) -> List[Path]:
    files = []

    for file_folder in files_folders:
        if file_folder.is_file() and file_folder.suffix == ".py":
            files.append(file_folder)
        elif file_folder.is_dir():
            files.extend(list(file_folder.glob("**/*.py")))
            files.extend(list(file_folder.glob("**/*.pyi")))

    return files


def _default_files_folders(app_context: AppContext[CorePluginConfig]) -> Tuple[Path, ...]:
    return (
        app_context.plugin_config.sources_directory,
        app_context.plugin_config.tests_directory,
        *[folder for folder in app_context.pyproject_toml.tool.delfino.local_command_folders if folder.exists()],
    )


class FormatSpinner(Spinner):
    def print_results_with_help(
        self,
        result: CompletedProcess,
        app_context: AppContext[CorePluginConfig],
        check: bool,
        msg: str,
        error_msg_match: str = "",
    ):
        """Prints the result of a command run.

        The command must be run with `stdout=PIPE` and `stderr=PIPE` to capture the output
        and show it after the command has finished.

        Args:
            result: The result of the command.
            error_msg_match: A string to match against the stdout and stderr. If any of the
                streams contains this string, the command is considered to have failed.
            app_context: The app context.
            check: Whether the command was run in check mode.
            msg: The message to show if the command failed.
        """
        if result.returncode > 0 or (
            error_msg_match and (error_msg_match in result.stdout.decode() or error_msg_match in result.stderr.decode())
        ):
            if check:  # if checking, hard fail
                msg_lines = [
                    f"{msg} before commit. Try following:",
                    f" * Run formatter manually with `{run_command_example(run_group_format, app_context)}` "
                    f"before committing code.",
                ]
                if not app_context.plugin_config.disable_pre_commit:
                    msg_lines.insert(
                        1,
                        " * Enable pre-commit hook by running `pre-commit install` in the repository.",
                    )

                self._print_failure(result)
                raise click.Abort()

            # if not checking, just print the error and continue
            self._print_failure(result)
            return

        self._print_success()


@click.command("pyupgrade")
@files_folders_option
@pass_args
@pass_plugin_app_context
def run_pyupgrade(
    app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[Path, ...], **kwargs
):
    """Runs pyupgrade with automatic version discovery.

    The Python version comes from the `pyproject.toml` file.
    """
    assert_pip_package_installed("pyupgrade")
    args = list(passed_args)

    if kwargs.get("check", False):
        args.append("--exit-zero-even-if-changed")

    python_version = app_context.pyproject_toml.tool.poetry.dependencies.get("python", "3")
    # see `pypgrade -h` for range of supported versions
    min_python_version = _find_min_minor_version(python_version, 3, 6, 11)
    args.append(f"--py3{min_python_version}-plus")

    spinner = FormatSpinner(
        "pyupgrade", f"upgrading code to Python 3{'.' + min_python_version if min_python_version else ''}"
    )
    results = run(
        ["pyupgrade", *_all_python_files(files_folders or _default_files_folders(app_context)), *args],
        stdout=PIPE,
        stderr=PIPE,
        on_error=OnError.PASS,
        running_hook=spinner,
    )
    spinner.print_results_with_help(
        results,
        app_context,
        bool("--exit-zero-even-if-changed" in args),
        f"Code was not upgraded to Python 3{'.' + min_python_version if min_python_version else ''}",
    )


@click.command("black")
@files_folders_option
@pass_args
@pass_plugin_app_context
def run_black(
    app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[Path, ...], **kwargs
):
    """Runs black."""
    assert_pip_package_installed("black")
    args = list(passed_args)

    if kwargs.get("check", False):
        args.append("--check")

    if kwargs.get("quiet", False):
        args.append("--quiet")

    spinner = FormatSpinner("black", "formatting code")

    results = run(
        ["black", *args, *(files_folders or _default_files_folders(app_context))],
        stdout=PIPE,
        stderr=PIPE,
        on_error=OnError.PASS,
        running_hook=spinner,
    )

    spinner.print_results_with_help(
        results, app_context, bool("--check" in args), "Code was not formatted", "reformatted "
    )


@click.command("isort")
@files_folders_option
@pass_args
@pass_plugin_app_context
def run_isort(
    app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[Path, ...], **kwargs
):
    """Runs isort."""
    assert_pip_package_installed("isort")
    args = list(passed_args)

    if kwargs.get("check", False):
        args.append("--check")

    if kwargs.get("quiet", False):
        args.append("--quiet")

    spinner = FormatSpinner("isort", "sorting imports")
    results = run(
        ["isort", *args, *(files_folders or _default_files_folders(app_context))],
        stdout=PIPE,
        stderr=PIPE,
        on_error=OnError.PASS,
        running_hook=spinner,
    )

    spinner.print_results_with_help(results, app_context, bool("--check" in args), "Imports were not sorted")


@click.command("ensure-pre-commit")
@pass_plugin_app_context
def run_ensure_pre_commit(app_context: AppContext[CorePluginConfig], **kwargs):
    """Ensures pre-commit is installed and enabled."""
    del kwargs  # not used, needed for extra kwargs from the command group
    if not app_context.plugin_config.disable_pre_commit:
        assert_pip_package_installed("pre-commit")
        # ensure pre-commit is installed
        run("pre-commit install", stdout=PIPE, on_error=OnError.EXIT)


@click.command("format", help=commands_group_help("format"))
@click.option("--check", is_flag=True, help="Only check formatting, don't reformat the code.")
@click.option("--quiet", is_flag=True, help="Don't show progress. Only errors.")
@files_folders_option
@pass_plugin_app_context
@click.pass_context
def run_group_format(
    click_context: click.Context,
    app_context: AppContext[CorePluginConfig],
    files_folders: Tuple[Path, ...],
    check: bool,
    quiet: bool,
):
    execute_commands_group(
        click_context,
        app_context.plugin_config,
        files_folders=files_folders,
        check=check,
        quiet=quiet,
    )
